# libraries
import warnings
import traceback
import pandas as pd
import geopandas as gpd
from shapely import wkt
import multiprocessing as mp
from multiprocessing.pool import ThreadPool
import time
import progressbar as pb
import logging
import itertools
import datetime
from sqlalchemy import text as sqltxt
import sqlalchemy as sa
import textwrap

from ..db.connections import db_engine
from ..utils.dwd import get_dwd_meta, get_cdc_file_list
from ..station.StationBases import StationBase
from ..db import models
from ..db.queries.get_quotient import _get_quotient

# set settings
# ############
try:# else I get strange errors with linux
    mp.set_start_method('spawn')
except RuntimeError:
    pass

__all__ = ["StationsBase"]
log = logging.getLogger(__name__)

# Base class definitions
########################

class StationsBase:
    _StationClass = StationBase
    _timeout_raw_imp = 240

    def __init__(self):
        if type(self) is StationsBase:
            raise NotImplementedError("""
            The StationsBase is only a wrapper class an is not working on its own.
            Please use StationP, StationPD, StationT or StationET instead""")
        self._ftp_folder_base = self._StationClass._ftp_folder_base
        if isinstance(self._ftp_folder_base, str):
            self._ftp_folder_base = [self._ftp_folder_base]

        # create ftp_folders in order of importance
        self._ftp_folders = list(itertools.chain(*[
            [base + "historical/", base + "recent/"]
                for base in self._ftp_folder_base]))

        self._para = self._StationClass._para
        self._para_long = self._StationClass._para_long

    def download_meta(self):
        """Download the meta file(s) from the CDC server.

        Returns
        -------
        geopandas.GeoDataFrame
            The meta file from the CDC server.
            If there are several meta files on the server, they are joined together.
        """
        # download historic meta file
        meta = get_dwd_meta(self._ftp_folders[0])

        for ftp_folder in self._ftp_folders[1:]:
            meta_new = get_dwd_meta(ftp_folder=ftp_folder)

            # add new stations
            meta = pd.concat(
                [meta, meta_new[~meta_new.index.isin(meta.index)]])
            if isinstance(meta_new, gpd.GeoDataFrame):
                meta = gpd.GeoDataFrame(meta, crs=meta_new.crs)

            # check for wider timespan
            if "bis_datum" in meta.columns:
                meta = meta.join(
                    meta_new[["bis_datum", "von_datum"]],
                    how="left", rsuffix="_new")

                mask = meta["von_datum"] > meta["von_datum_new"]
                meta.loc[mask, "von_datum"] = meta_new.loc[mask, "von_datum"]

                mask = meta["bis_datum"] < meta["bis_datum_new"]
                meta.loc[mask, "bis_datum"] = meta_new.loc[mask, "bis_datum"]

                meta.drop(["von_datum_new", "bis_datum_new"], axis=1, inplace=True)

        return meta

    @db_engine.deco_update_privilege
    def update_meta(self, stids="all", **kwargs):
        """Update the meta table by comparing to the CDC server.

        The "von_datum" and "bis_datum" is ignored because it is better to set this by the filled period of the stations in the database.
        Often the CDC period is not correct.

        Parameters
        ----------
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        """
        log.info(
            "The {para_long} meta table gets updated."\
                .format(para_long=self._para_long))
        meta = self.download_meta()

        # check if Abgabe is in meta
        if "Abgabe" in meta.columns:
            meta.drop("Abgabe", axis=1, inplace=True)

        # get droped stations and delete from meta file
        sql_get_droped = sa\
            .select(models.DropedStations.station_id)\
            .where(models.DropedStations.parameter == self._para)
        with db_engine.connect() as con:
            droped_stids = con.execute(sql_get_droped).all()
        droped_stids = [row[0] for row in droped_stids
                        if row[0] in meta.index]
        meta.drop(droped_stids, inplace=True)

        # check if only some stids should be updated
        if stids != "all":
            if not isinstance(stids, list):
                stids = [stids,]
            meta.drop([stid for stid in meta.index if stid not in stids], inplace=True)

        # to have a meta entry for every station before looping over them
        if "von_datum" in meta.columns and "bis_datum" in meta.columns:
            self._update_db_meta(
                meta=meta.drop(["von_datum", "bis_datum"], axis=1))
        else:
            self._update_db_meta(meta=meta)

        log.info(
            "The {para_long} meta table got successfully updated."\
                .format(para_long=self._para_long))

    @db_engine.deco_update_privilege
    def _update_db_meta(self, meta):
        """Update a meta table on the database with new DataFrame.

        Parameters
        ----------
        meta : pandas.DataFrame
            A DataFrame with station_id as index.
        """
        # get the columns of meta
        meta = meta.rename_axis("station_id").reset_index()
        columns = [col.lower() for col in meta.columns]
        columns = columns + ['geometry_utm'] if 'geometry' in columns else columns
        meta.rename(dict(zip(meta.columns, columns)), axis=1, inplace=True)

        # check if columns are initiated in DB
        with db_engine.connect() as con:
            columns_db = con.execute(sqltxt(
                """
                SELECT column_name
                FROM information_schema.columns
                WHERE table_name='meta_{para}';
                """.format(para=self._para)
                )).all()
            columns_db = [col[0] for col in columns_db]

        problem_cols = [col for col in columns if col not in columns_db]
        if len(problem_cols) > 0:
            warnings.warn("""
                The meta_{para} column '{cols}' is not initiated in the database.
                This column is therefor skiped.
                Please review the DB or the code.
                """.format(
                    para=self._para,
                    cols=", ".join(problem_cols))
                    )
            columns = [col for col in columns if col in columns_db]

        # change date columns
        for colname, col in \
                meta.select_dtypes(include="datetime64").items():
            meta.loc[:,colname] = col.dt.strftime("%Y%m%d %H:%M")

        # change geometry
        if "geometry" in meta.columns:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                meta["geometry_utm"] = meta.geometry.to_crs(25832).to_wkt()
                meta["geometry"] = meta.geometry.to_crs(4326).to_wkt()

        # change all to strings
        meta = meta.astype(str)

        # get values
        values_all = ["', '".join(pair) for pair in meta.loc[:,columns].values]
        values = "('" + "'), ('".join(values_all) + "')"
        values = values.replace("'nan'", "NULL").replace("'<NA>'", "NULL")

        # create sql
        sql = '''
            INSERT INTO meta_{para}({columns})
            Values {values}
            ON CONFLICT (station_id) DO UPDATE SET
            '''.format(
                columns=", ".join(columns),
                values=values,
                para=self._para)
        for col in columns:
            sql += ' "{col}" = EXCLUDED."{col}", '.format(col=col)

        sql = sql[:-2] + ";"

        # run sql command
        with db_engine.connect() as con:
            con.execute(sqltxt(sql))
            con.commit()

    @db_engine.deco_update_privilege
    def update_period_meta(self, stids="all", **kwargs):
        """Update the period in the meta table of the raw data.

        Parameters
        ----------
        stids: string  or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs : dict, optional
        **kwargs : dict, optional
            The additional keyword arguments are passed to the get_stations method.

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        self._run_simple_loop(
            stations=self.get_stations(only_real=True, stids=stids, **kwargs),
            method="update_period_meta",
            name="update period in meta",
            kwargs=kwargs
        )

    @classmethod
    def get_meta_explanation(cls, infos="all"):
        """Get the explanations of the available meta fields.

        Parameters
        ----------
        infos : list or string, optional
            The infos you wish to get an explanation for.
            If "all" then all the available information get returned.
            The default is "all"

        Returns
        -------
        pd.Series
            a pandas Series with the information names as index and the explanation as values.
        """
        return cls._StationClass.get_meta_explanation(infos=infos)

    def get_meta(self,
            infos=["station_id", "filled_from", "filled_until", "geometry"],
            stids="all",
            only_real=True):
        """Get the meta Dataframe from the Database.

        Parameters
        ----------
        infos : list or str, optional
            A list of information from the meta file to return
            If "all" than all possible columns are returned, but only one geometry column.
            The default is: ["Station_id", "filled_from", "filled_until", "geometry"]
        only_real: bool, optional
            Whether only real stations are returned or also virtual ones.
            True: only stations with own data are returned.
            The default is True.

        Returns
        -------
        pandas.DataFrame or geopandas.GeoDataFrae
            The meta DataFrame.
        """
        # make sure columns is of type list
        if isinstance(infos, str):
            if infos=="all":
                infos = self.get_meta_explanation(infos="all").index.to_list()
                if "geometry_utm" in infos:
                    infos.remove("geometry_utm")
            else:
                infos = [infos]

        # check infos
        infos = [col.lower() for col in infos]
        if "station_id" not in infos:
            infos.insert(0, "station_id")
        if "geometry" in infos and "geometry_utm" in infos:
            warnings.warn(textwrap.dedent("""\
                You selected 2 geometry columns.
                Only the geometry column with EPSG 4326 is returned"""))
            infos.remove("geometry_utm")

        # create geometry select statement
        infos_select = []
        for info in infos:
            if info in ["geometry", "geometry_utm"]:
                infos_select.append(
                    f"ST_AsText({info}) as {info}")
            else:
                infos_select.append(info)

        # create sql statement
        sql = "SELECT {cols} FROM meta_{para}"\
            .format(cols=", ".join(infos_select), para=self._para)
        if only_real:
            where_clause = " WHERE is_real=true"
        if stids != "all":
            if not isinstance(stids, list):
                stids = [stids,]
            if "where_clause" not in locals():
                where_clause = " WHERE "
            else:
                where_clause += " AND "
            where_clause += "station_id in ({stids})".format(
                stids=", ".join([str(stid) for stid in stids]))
        if "where_clause" in locals():
            sql += where_clause

        # execute queries to db
        with db_engine.connect() as con:
            meta = pd.read_sql(
                sqltxt(sql),
                con,
                index_col="station_id")

        # make datetime columns timezone aware
        meta = meta.apply(
            lambda col: col.dt.tz_localize(datetime.timezone.utc) \
                if hasattr(col, "dt") and not col.dt.tz else col)

        # change to GeoDataFrame if geometry column was selected
        for geom_col, srid in zip(["geometry", "geometry_utm"],
                                  ["4326",     "25832"]):
            if geom_col in infos:
                meta[geom_col] = meta[geom_col].apply(wkt.loads)
                meta = gpd.GeoDataFrame(
                    meta, crs="EPSG:" + srid, geometry=geom_col)

        # strip whitespaces in string columns
        for col in meta.columns[meta.dtypes == "object"]:
            try:
                meta[col] = meta[col].str.strip()
            except:
                pass

        return meta

    def get_stations(self, only_real=True, stids="all", skip_missing_stids=False, **kwargs):
        """Get a list with all the stations as Station-objects.

        Parameters
        ----------
        only_real: bool, optional
            Whether only real stations are returned or also virtual ones.
            True: only stations with own data are returned.
            The default is True.
        stids: string or list of int, optional
            The Stations to return.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        skip_missing_stids: bool, optional
            Should the method skip the missing stations from input stids?
            If False, then a ValueError is raised if a station is not found.
            The default is False.
        **kwargs : dict, optional
            The additional keyword arguments aren't used in this method.

        Returns
        -------
        Station-object
            returns a list with the corresponding station objects.

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        meta = self.get_meta(
            infos=["station_id"], only_real=only_real, stids=stids)

        if isinstance(stids, str) and (stids == "all"):
            stations = [
                self._StationClass(stid, _skip_meta_check=True)
                for stid in meta.index]
        else:
            stids = list(stids)
            stations = [
                self._StationClass(stid, _skip_meta_check=True)
                for stid in meta.index
                if stid in stids]
            if (not skip_missing_stids) and (len(stations) != len(stids)):
                stations_ids = [stat.id for stat in stations]
                raise ValueError(
                    "It was not possible to create a {para_long} Station with the following IDs: {stids}".format(
                        para_long=self._para_long,
                        stids = ", ".join([str(stid) for stid in stids if stid not in stations_ids])
                    ))

        return stations

    def get_quotient(self, kinds_num, kinds_denom, stids="all", return_as="df",  **kwargs):
        """Get the quotient of multi-annual means of two different kinds or the timeserie and the multi annual raster value.

        $quotient = \overline{ts}_{kind_num} / \overline{ts}_{denom}$

        Parameters
        ----------
        kinds_num : list of str or str
            The timeseries kinds of the numerators.
            Should be one of ['raw', 'qc', 'filled'].
            For precipitation also "corr" is possible.
        kinds_denom : list of str or str
            The timeseries kinds of the denominator or the multi annual raster key.
            If the denominator is a multi annual raster key, then the result is the quotient of the timeserie and the raster value.
            Possible values are:
                - for timeserie kinds: 'raw', 'qc', 'filled' or for precipitation also "corr".
                - for raster keys: 'hyras', 'dwd' or 'regnie', depending on your defined raster files.
        stids : list of Integer
            The stations IDs for which to compute the quotient.
        return_as : str, optional
            The format of the return value.
            If "df" then a pandas DataFrame is returned.
            If "json" then a list with dictionaries is returned.
        **kwargs : dict, optional
            The additional keyword arguments are passed to the get_stations method.

        Returns
        -------
        pandas.DataFrame or list of dict
            The quotient of the two timeseries as DataFrame or list of dictionaries (JSON) depending on the return_as parameter.
            The default is pd.DataFrame.

        Raises
        ------
        ValueError
            If the input parameters were not correct.
        """
        # check stids
        if stids == "all":
            stids = None

        # check kinds
        rast_keys = {"hyras", "regnie", "dwd"}
        kinds_num = self._StationClass._check_kinds(kinds_num)
        kinds_denom = self._StationClass._check_kinds(
            kinds_denom,
            valids=self._StationClass._valid_kinds | rast_keys)

        # get quotient
        with db_engine.connect() as con:
            return _get_quotient(
                con=con,
                stids=stids,
                paras=self._para,
                kinds_num=kinds_num,
                kinds_denom=kinds_denom,
                return_as=return_as)

    def count_holes(self, stids="all", **kwargs):
        """Count holes in timeseries depending on there length.

        Parameters
        ----------
        stids: string or list of int, optional
            The Stations to return.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs : dict, optional
        **kwargs : dict, optional
            This is a list of parameters, that is supported by the StationBase.count_holes method.

            Furthermore the kwargs are passed to the get_stations method.

            possible values are:

            - weeks : list, optional
                A list of hole length to count.
                Every hole longer than the duration of weeks specified is counted.
                The default is [2, 4, 8, 12, 16, 20, 24]
            - kind : str
                The kind of the timeserie to analyze.
                Should be one of ['raw', 'qc', 'filled'].
                For N also "corr" is possible.
                Normally only "raw" and "qc" make sense, because the other timeseries should not have holes.
            - period : TimestampPeriod or (tuple or list of datetime.datetime or None), optional
                The minimum and maximum Timestamp for which to analyze the timeseries.
                If None is given, the maximum and minimal possible Timestamp is taken.
                The default is (None, None).
            - between_meta_period : bool, optional
                Only check between the respective period that is defined in the meta table.
                If "qc" is chosen as kind, then the "raw" meta period is taken.
                The default is True.
            - crop_period : bool, optional
                should the period get cropped to the maximum filled period.
                This will result in holes being ignored when they are at the end or at the beginning of the timeserie.
                If period = (None, None) is given, then this parameter is set to True.
                The default is False.

        Returns
        -------
        pandas.DataFrame
            A Pandas Dataframe, with station_id as index and one column per week.
            The numbers in the table are the amount of NA-periods longer than the respective amount of weeks.

        Raises
        ------
        ValueError
            If the input parameters were not correct.
        """
        # check input parameters
        stations = self.get_stations(stids=stids, only_real=True, **kwargs)

        # iter stations
        first = True
        for station in pb.progressbar(stations, line_breaks=False):
            new_count = station.count_holes(**kwargs)
            if first:
                meta = new_count
                first = False
            else:
                meta = pd.concat([meta, new_count], axis=0)

        return meta

    @staticmethod
    def _get_progressbar(max_value, name):
        pbar = pb.ProgressBar(
            widgets=[
                pb.widgets.RotatingMarker(),
                " " + name ,
                pb.widgets.Percentage(), ' ',
                pb.widgets.SimpleProgress(
                    format=("('%(value_s)s/%(max_value_s)s')")), ' ',
                pb.widgets.Bar(min_width=80), ' ',
                pb.widgets.Timer(format='%(elapsed)s'), ' | ',
                pb.widgets.ETA(),
                pb.widgets.DynamicMessage(
                    "last_station",
                    format=", last id: {formatted_value}",
                    precision=4)
                ],
            max_value=max_value,
            variables={"last_station": "None"},
            term_width=100,
            is_terminal=True
            )
        pbar.update(0)

        return pbar

    def _run_method(self, stations, method, name, kwds=dict(),
            do_mp=True, processes=mp.cpu_count()-1, **kwargs):
        """Run methods of the given stations objects in multiprocessing/threading mode.

        Parameters
        ----------
        stations : list of station objects
            A list of station objects. Those must be children of the StationBase class.
        method : str
            The name of the method to call.
        name : str
            A descriptive name of the method to show in the progressbar.
        kwds : dict
            The keyword arguments to give to the methods
        do_mp : bool, optional
            Should the method be done in multiprocessing mode?
            If False the methods will be called in threading mode.
            Multiprocessing needs more memory and a bit more initiating time. Therefor it is only usefull for methods with a lot of computation effort in the python code.
            If the most computation of a method is done in the postgresql database, then threading is enough to speed the process up.
            The default is True.
        processes : int, optional
            The number of processes that should get started simultaneously.
            If 1 or less, then the process is computed as a simple loop, so there is no multiprocessing or threading done.
            The default is the cpu count -1.
        """
        log.info(
            f"{self._para_long} Stations async loop over method '{method}' started." +
            "\n" +"-"*80
            )

        if processes<=1:
            log.info(f"As the number of processes is 1 or lower, the method '{method}' is started as a simple loop.")
            self._run_simple_loop(
                stations=stations, method=method, name=name, kwds=kwds)
        else:
            # progressbar
            num_stations = len(stations)
            pbar = self._get_progressbar(max_value=num_stations, name=name)

            # create pool
            if do_mp:
                try:
                    pool = mp.Pool(processes=processes)
                    log.debug("the multiprocessing Pool is started")
                except AssertionError:
                    log.debug('daemonic processes are not allowed to have children, therefor threads are used')
                    pool = ThreadPool(processes=processes)
            else:
                log.debug("the threading Pool is started")
                pool = ThreadPool(processes=processes)

            # start processes
            results = []
            for stat in stations:
                results.append(
                    pool.apply_async(
                        getattr(stat, method),
                        kwds=kwds))
            pool.close()

            # check results until all finished
            finished = [False] * num_stations
            while (True):
                if all(finished):
                    break

                for result in [result for i, result in enumerate(results)
                                    if not finished[i] and result.ready()]:
                    index = results.index(result)
                    finished[index] = True
                    pbar.variables["last_station"] = stations[index].id
                    # get stdout and log
                    header = f"""The {name} of the {self._para_long} Station with ID {stations[index].id} finished with """
                    try:
                        stdout = result.get(10)
                        if stdout is not None:
                            log.debug(f"{header}stdout:\n{result.get(10)}")
                    except Exception:
                        log.error(f"{header}stderr:\n{traceback.format_exc()}")

                    pbar.update(sum(finished))
                time.sleep(2)

            pbar.update(sum(finished))
            pool.join()
            pool.terminate()

    def _run_simple_loop(self, stations, method, name, kwds=dict()):
        log.info("-"*79 +
        "\n{para_long} Stations simple loop over method '{method}' started.".format(
            para_long=self._para_long,
            method=method
        ))

        # progressbar
        num_stations = len(stations)
        pbar = self._get_progressbar(max_value=num_stations, name=name)

        # start processes
        for stat in stations:
            getattr(stat, method)(**kwds)
            pbar.variables["last_station"] = stat.id
            pbar.update(pbar.value + 1)

    @db_engine.deco_update_privilege
    def update_raw(self, only_new=True, only_real=True, stids="all",
            remove_nas=True, do_mp=True, **kwargs):
        """Download all stations data from CDC and upload to database.

        Parameters
        ----------
        only_new : bool, optional
            Get only the files that are not yet in the database?
            If False all the available files are loaded again.
            The default is True
        only_real: bool, optional
            Whether only real stations are tried to download.
            True: only stations with a date in raw_from in meta are downloaded.
            The default is True.
        stids: string or list of int, optional
            The Stations to return.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        do_mp : bool, optional
            Should the method be done in multiprocessing mode?
            If False the methods will be called in threading mode.
            Multiprocessing needs more memory and a bit more initiating time. Therefor it is only usefull for methods with a lot of computation effort in the python code.
            If the most computation of a method is done in the postgresql database, then threading is enough to speed the process up.
            The default is True.
        remove_nas : bool, optional
            Remove the NAs from the downloaded data before updating it to the database.
            This has computational advantages.
            The default is True.
        **kwargs : dict, optional
            The additional keyword arguments for the _run_method and get_stations method

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        start_tstp = datetime.datetime.now()

        # get FTP file list
        ftp_file_list = get_cdc_file_list(
            ftp_folders=self._ftp_folders)

        # run the tasks in multiprocessing mode
        self._run_method(
            stations=self.get_stations(only_real=only_real, stids=stids, **kwargs),
            method="update_raw",
            name="download raw {para} data".format(para=self._para.upper()),
            kwds=dict(
                only_new=only_new,
                ftp_file_list=ftp_file_list,
                remove_nas=remove_nas),
            do_mp=do_mp, **kwargs)

        # save start time as variable to db
        do_update_period = isinstance(stids, str) and (stids == "all")
        if not do_update_period and isinstance(stids, list):
            all_stids = self.get_meta(["station_id"], stids="all", only_real=True).index
            do_update_period = all([stid in stids for stid in all_stids])

        if do_update_period:
            with db_engine.connect() as con:
                con.execute(sqltxt("""
                    INSERT INTO parameter_variables (parameter, start_tstp_last_imp, max_tstp_last_imp)
                    VALUES ('{para}',
                            '{start_tstp}'::timestamp,
                            (SELECT max(raw_until) FROM meta_{para}))
                    ON CONFLICT (parameter) DO UPDATE SET
                        start_tstp_last_imp=EXCLUDED.start_tstp_last_imp,
                        max_tstp_last_imp=EXCLUDED.max_tstp_last_imp;
                """.format(
                    para=self._para,
                    start_tstp=start_tstp.strftime("%Y%m%d %H:%M"))))
                con.commit()

    @db_engine.deco_update_privilege
    def last_imp_quality_check(self, stids="all", do_mp=False, **kwargs):
        """Do the quality check of the last import.

        Parameters
        ----------
        do_mp : bool, optional
            Should the method be done in multiprocessing mode?
            If False the methods will be called in threading mode.
            Multiprocessing needs more memory and a bit more initiating time. Therefor it is only usefull for methods with a lot of computation effort in the python code.
            If the most computation of a method is done in the postgresql database, then threading is enough to speed the process up.
            The default is False.
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs : dict, optional
            The additional keyword arguments for the _run_method and get_stations method
        """
        self._run_method(
            stations=self.get_stations(only_real=True, stids=stids, **kwargs),
            method="last_imp_quality_check",
            name="quality check {para} data".format(para=self._para.upper()),
            do_mp=do_mp, **kwargs)

    @db_engine.deco_update_privilege
    def last_imp_fillup(self, stids="all", do_mp=False, **kwargs):
        """Do the gap filling of the last import.

        Parameters
        ----------
        do_mp : bool, optional
            Should the method be done in multiprocessing mode?
            If False the methods will be called in threading mode.
            Multiprocessing needs more memory and a bit more initiating time. Therefor it is only usefull for methods with a lot of computation effort in the python code.
            If the most computation of a method is done in the postgresql database, then threading is enough to speed the process up.
            The default is False.
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs : dict, optional
            The additional keyword arguments for the _run_method and get_stations method
        """
        stations = self.get_stations(only_real=False, stids=stids, **kwargs)
        period = stations[0].get_last_imp_period(all=True)
        period_log = period.strftime("%Y-%m-%d %H:%M")
        log.info("The {para_long} Stations fillup of the last import is started for the period {min_tstp} - {max_tstp}".format(
            para_long=self._para_long,
            min_tstp=period_log[0],
            max_tstp=period_log[1]))
        self._run_method(
            stations=stations,
            method="last_imp_fillup",
            name="fillup {para} data".format(para=self._para.upper()),
            kwds=dict(_last_imp_period=period),
            do_mp=do_mp,
            **kwargs)

    @db_engine.deco_update_privilege
    def quality_check(self, period=(None, None), only_real=True, stids="all",
            do_mp=False, **kwargs):
        """Quality check the raw data for a given period.

        Parameters
        ----------
        period : tuple or list of datetime.datetime or None, optional
            The minimum and maximum Timestamp for which to get the timeseries.
            If None is given, the maximum or minimal possible Timestamp is taken.
            The default is (None, None).
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        do_mp : bool, optional
            Should the method be done in multiprocessing mode?
            If False the methods will be called in threading mode.
            Multiprocessing needs more memory and a bit more initiating time. Therefor it is only usefull for methods with a lot of computation effort in the python code.
            If the most computation of a method is done in the postgresql database, then threading is enough to speed the process up.
            The default is False.
        **kwargs : dict, optional
            The additional keyword arguments for the _run_method and get_stations method
        """
        self._run_method(
            stations=self.get_stations(only_real=only_real, stids=stids, **kwargs),
            method="quality_check",
            name="quality check {para} data".format(para=self._para.upper()),
            kwds=dict(period=period),
            do_mp=do_mp,
            **kwargs)

    @db_engine.deco_update_privilege
    def update_ma_raster(self, stids="all", do_mp=False, **kwargs):
        """Update the multi annual values for the stations.

        Get a multi annual value from the corresponding raster and save to the multi annual table in the database.

        Parameters
        ----------
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        do_mp : bool, optional
            Should the method be done in multiprocessing mode?
            If False the methods will be called in threading mode.
            Multiprocessing needs more memory and a bit more initiating time. Therefor it is only usefull for methods with a lot of computation effort in the python code.
            If the most computation of a method is done in the postgresql database, then threading is enough to speed the process up.
            The default is False.
        **kwargs : dict, optional
            The additional keyword arguments for the _run_method and get_stations method

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        self._run_method(
            stations=self.get_stations(only_real=False, stids=stids, **kwargs),
            method="update_ma_raster",
            name="update ma-values for {para}".format(para=self._para.upper()),
            do_mp=do_mp,
            **kwargs)

    @db_engine.deco_update_privilege
    def fillup(self, only_real=False, stids="all", do_mp=False, **kwargs):
        """Fill up the quality checked data with data from nearby stations to get complete timeseries.

        Parameters
        ----------
        only_real: bool, optional
            Whether only real stations are computed or also virtual ones.
            True: only stations with own data are returned.
            The default is True.
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        do_mp : bool, optional
            Should the method be done in multiprocessing mode?
            If False the methods will be called in threading mode.
            Multiprocessing needs more memory and a bit more initiating time. Therefor it is only usefull for methods with a lot of computation effort in the python code.
            If the most computation of a method is done in the postgresql database, then threading is enough to speed the process up.
            The default is False.
        **kwargs : dict, optional
            The additional keyword arguments for the _run_method and get_stations method

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        self._run_method(
            stations=self.get_stations(only_real=only_real, stids=stids, **kwargs),
            method="fillup",
            name="fillup {para} data".format(para=self._para.upper()),
            do_mp=do_mp,
            **kwargs)

    @db_engine.deco_update_privilege
    def update(self, only_new=True, **kwargs):
        """Make a complete update of the stations.

        Does the update_raw, quality check and fillup of the stations.

        Parameters
        ----------
        only_new : bool, optional
            Should a only new values be computed?
            If False: The stations are updated for the whole possible period.
            If True, the stations are only updated for new values.
            The default is True.
        """
        self.update_raw(only_new=only_new, **kwargs)
        if only_new:
            self.last_imp_quality_check(**kwargs)
            self.last_imp_fillup(**kwargs)
        else:
            self.quality_check(**kwargs)
            self.fillup(**kwargs)

    def get_df(self, stids, **kwargs):
        """Get a DataFrame with the corresponding data.

        Parameters
        ----------
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs: optional keyword arguments
            Those keyword arguments are passed to the get_df function of the station class.
            Possible parameters are period, agg_to, kinds.
            Furthermore the kwargs are passed to the get_stations method.

        Returns
        -------
        pd.Dataframe
            A DataFrame with the timeseries for the selected stations, kind(s) and the given period.
            If multiple columns are selected, the columns in this DataFrame is a MultiIndex with the station IDs as first level and the kind as second level.
        """
        if "kinds" in kwargs and "kind" in kwargs:
            raise ValueError("Either enter kind or kinds, not both.")
        if "kind" in kwargs:
            kinds=[kwargs.pop("kind")]
        else:
            kinds=kwargs.pop("kinds")
        kwargs.update(dict(only_real=kwargs.get("only_real", False)))
        stats = self.get_stations(stids=stids, **kwargs)
        df_all = None
        for stat in pb.progressbar(stats, line_breaks=False):
            df = stat.get_df(kinds=kinds, **kwargs)
            if df is None:
                warnings.warn(
                    f"There was no data for {stat._para_long} station {stat.id}!")
                continue
            if len(df.columns) == 1:
                df.rename(
                    dict(zip(df.columns, [stat.id])),
                    axis=1, inplace=True)
            else:
                df.columns = pd.MultiIndex.from_product(
                    [[stat.id], df.columns],
                    names=["Station ID", "kind"])
            df_all = pd.concat([df_all, df], axis=1)

        return df_all



import pandas as pd
import sqlalchemy as sa
from ..views import StationKindQuotientView, StationMATimeserieQuotientView


def _get_quotient(con, stids, paras, kinds_num, kinds_denom, return_as):
    """Get the quotient of multi-annual means of two different kinds or the timeserie and the multi annual raster value.

    $quotient = \overline{ts}_{kind_num} / \overline{ts}_{denom}$

    Parameters
    ----------
    con : sqlalchemy.engine.base.Connection
        The connection to the database.
    stids : list of int or int or None
        The station ids to get the quotient from.
        If None, then all stations in the database are used.
    paras : list of str or str
        The parameters to get the quotient from.
    kinds_num : list of str or str
        The timeseries kinds of the numerators.
        Should be one of ['raw', 'qc', 'filled'].
        For precipitation also "corr" is possible.
    kinds_denom : list of str or str
        The timeseries kinds of the denominator or the multi annual raster key.
        If the denominator is a multi annual raster key, then the result is the quotient of the timeserie and the raster value.
        Possible values are:
            - for timeserie kinds: 'raw', 'qc', 'filled' or for precipitation also "corr".
            - for raster keys: 'hyras', 'dwd' or 'regnie', depending on your defined raster files.
    return_as : str
        The format of the return value.
        If "df" then a pandas DataFrame is returned.
        If "json" then a list with dictionaries is returned.

    Returns
    -------
    pandas.DataFrame or list of dict
        The quotient of the two timeseries as DataFrame or list of dictionaries (JSON) depending on the return_as parameter.
        The default is pd.DataFrame.

    Raises
    ------
    ValueError
        If the input parameters were not correct.
    """
    # split ts and raster kinds
    rast_keys = {"hyras", "regnie", "dwd"}
    kinds_denom_ts = set(kinds_denom) - rast_keys
    kinds_denom_rast = set(kinds_denom) & rast_keys

    # create tests
    tests_ts = []
    tests_rast = []

    # create stids tests
    if (isinstance(stids, list) or isinstance(stids, set)) and len(stids) == 1:
        stids = stids[0]
    if isinstance(stids, int):
        tests_ts.append(StationKindQuotientView.station_id == stids)
        tests_rast.append(StationMATimeserieQuotientView.station_id == stids)
    elif isinstance(stids, list) or isinstance(stids, set):
        tests_ts.append(StationKindQuotientView.station_id.in_(stids))
        tests_rast.append(StationMATimeserieQuotientView.station_id.in_(stids))
    elif stids is not None:
        raise ValueError("The stids parameter should be a list of integers or an integer or None.")

    # create paras tests
    if (isinstance(paras, list) or isinstance(paras, set)) and len(paras) == 1:
        paras = paras[0]
    if isinstance(paras, str):
        tests_ts.append(StationKindQuotientView.parameter == paras)
        tests_rast.append(StationMATimeserieQuotientView.parameter.like(f"{paras}%"))
    elif isinstance(paras, list) or isinstance(paras, set):
        tests_ts.append(StationKindQuotientView.parameter.in_(paras))
        ors = []
        for para in paras:
            ors.append(StationMATimeserieQuotientView.parameter.like(f"{para}%"))
        tests_rast.append(sa.or_(*ors))
    else:
        raise ValueError("The paras parameter should be a list of strings or a string.")

    # create kinds_num tests
    if (isinstance(kinds_num, list) or isinstance(kinds_num, set)) and len(kinds_num) == 1:
        kinds_num = kinds_num[0]
    if isinstance(kinds_num, str):
        tests_ts.append(StationKindQuotientView.kind_numerator == kinds_num)
        tests_rast.append(StationMATimeserieQuotientView.kind == kinds_num)
    elif isinstance(kinds_num, list) or isinstance(kinds_num, set):
        tests_ts.append(StationKindQuotientView.kind_numerator.in_(kinds_num))
        tests_rast.append(StationMATimeserieQuotientView.kind.in_(kinds_num))
    else:
        raise ValueError("The kinds_num parameter should be a list of strings or a string.")

    # create kinds_denom tests
    if (isinstance(kinds_denom, list) or isinstance(kinds_denom, set)) and len(kinds_denom) == 1:
        kinds_denom = kinds_denom[0]
    if isinstance(kinds_denom, str):
        tests_ts.append(StationKindQuotientView.kind_denominator == kinds_denom)
        tests_rast.append(StationMATimeserieQuotientView.raster_key == kinds_denom)
    elif isinstance(kinds_denom, list) or isinstance(kinds_denom, set):
        tests_ts.append(StationKindQuotientView.kind_denominator.in_(kinds_denom))
        tests_rast.append(StationMATimeserieQuotientView.raster_key.in_(kinds_denom))
    else:
        raise ValueError("The kinds_denom parameter should be a list of strings or a string.")

    # get the quotient from the database views
    data = []
    # get timeseries quotient
    if len(kinds_denom_ts) > 0:
        stmnt = sa\
            .select(
                StationKindQuotientView.station_id,
                StationKindQuotientView.parameter,
                StationKindQuotientView.kind_numerator,
                StationKindQuotientView.kind_denominator,
                StationKindQuotientView.value)\
            .select_from(StationKindQuotientView.__table__)\
            .where(sa.and_(*tests_ts))

        data = con.execute(stmnt).all()

    # get raster quotient
    if len(kinds_denom_rast) > 0:
        stmnt = sa.\
            select(
                StationMATimeserieQuotientView.station_id,
                StationMATimeserieQuotientView.parameter,
                StationMATimeserieQuotientView.kind.label("kind_numerator"),
                StationMATimeserieQuotientView.raster_key.label("kind_denominator"),
                StationMATimeserieQuotientView.value)\
            .select_from(
                StationMATimeserieQuotientView.__table__)\
            .where(sa.and_(*tests_rast))

        data = data + con.execute(stmnt).all()

    # create return value
    if not any(data):
        data = []
    if return_as == "json":
        return [
            dict(
                station_id=row[0],
                parameter=row[1],
                kind_num=row[2],
                kind_denom=row[3],
                value=row[4])
            for row in data]
    elif return_as == "df":
        return pd.DataFrame(
            data,
            columns=["station_id", "parameter", "kind_num", "kind_denom", "value"]
        ).set_index(["station_id", "parameter", "kind_num", "kind_denom"])
    else:
        raise ValueError("The return_as parameter was not correct. Use one of ['df', 'json'].")


"""
Some utilities functions to get data from the DWD-CDC server.

Based on `max_fun` package on https://github.com/maxschmi/max_fun
Created by Max Schmit, 2021
"""
# libraries
import dateutil
import ftplib
import pathlib
import geopandas as gpd
import pandas as pd
from zipfile import ZipFile
import re
from io import BytesIO, StringIO
import traceback
import logging
import time
import random

# DWD - CDC FTP Server
CDC_HOST = "opendata.dwd.de"

# logger
log = logging.getLogger(__name__)

# basic functions
# ----------------
def dwd_id_to_str(id):
    """
    Convert a station id to normal DWD format as str.

    Parameters
    ----------
    id : int or str
        The id of the station.

    Returns
    -------
    str
        string of normal DWD Station id.

    """
    return f"{id:0>5}"

def _dwd_date_parser(date_ser):
    """
    Parse the dates from a DWD table to datetime.

    Parameters
    ----------
    date_ser : pd.Series of str or str
        the string from the DWD table. e.g. "20200101" or "2020010112"

    Returns
    -------
    datetime.datetime
        The date as datetime.

    """
    if not isinstance(date_ser, pd.Series):
        raise ValueError("date_str must be a pd.Series of str")

    # test if list or single str
    char_num = len(date_ser.iloc[0])

    # parse to correct datetime
    if char_num == 8:
        return pd.to_datetime(date_ser, format='%Y%m%d')
    elif char_num == 10:
        return pd.to_datetime(date_ser, format='%Y%m%d%H')
    elif char_num == 12:
        return pd.to_datetime(date_ser, format='%Y%m%d%H%M')
    else:
        raise ValueError("there was an error while converting the following  to a correct datetime"+
                         date_ser.head())

# functions
# ---------
def get_ftp_file_list(ftp_conn, ftp_folders):
    """Get a list of files in the folders with their modification dates.

    Parameters
    ----------
    ftp_conn : ftplib.FTP
        Ftp connection.
    ftp_folders : list of str or pathlike object
        The directories on the ftp server to look for files.

    Returns
    -------
    list of tuples of strs
        A list of Tuples. Every tuple stands for one file.
        The tuple consists of (filepath, modification date).
    """
    # check types
    if isinstance(ftp_folders, str):
        ftp_folders = [ftp_folders]
    for i, ftp_folder in enumerate(ftp_folders):
        if isinstance(ftp_folder, pathlib.Path):
            ftp_folders[i] = ftp_folder.as_posix()

    try:
        ftp_conn.voidcmd("NOOP")
    except ftplib.all_errors:
        ftp_conn.connect()

    # get files and modification dates
    files = []
    for ftp_folder in ftp_folders:
        lines = []
        ftp_conn.dir(ftp_folder, lines.append)
        for line in lines:
            parts = line.split(maxsplit=9)
            filepath = ftp_folder + parts[8]
            modtime = dateutil.parser.parse(parts[5] + " " + parts[6] + " " + parts[7])
            files.append((filepath, modtime))

    return files

def get_cdc_file_list(ftp_folders):
    with ftplib.FTP(CDC_HOST) as ftp_con:
        ftp_con.login()
        files = get_ftp_file_list(ftp_con, ftp_folders)
    return files

def get_dwd_file(zip_filepath):
    """
    Get a DataFrame from one single (zip-)file from the DWD FTP server.

    Parameters
    ----------
    zip_filepath : str
        Path to the file on the server. e.g.
        - "/climate_environment/CDC/observations_germany/climate/10_minutes/air_temperature/recent/10minutenwerte_TU_00044_akt.zip"
        - "/climate_environment/CDC/derived_germany/soil/daily/historical/derived_germany_soil_daily_historical_73.txt.gz"

    Returns
    -------
    pandas.DataFrame
        The DataFrame of the selected file in the zip folder.

    """
    # get the compressed folder from dwd
    with ftplib.FTP(CDC_HOST) as ftp:
        ftp.login()

        # download file
        compressed_bin = BytesIO()
        num_tried = 0
        while num_tried < 10:
            try:
                ftp.retrbinary("RETR " + zip_filepath, compressed_bin.write)
                break
            except Exception as e:
                if num_tried < 9:
                    num_tried += 1
                    time.sleep(random.randint(0,400)/100)
                else:
                    raise e

    # check folder to be derived or observation type import the data
    if re.search("observations", zip_filepath):
        # get zip folder and files
        compressed_folder = ZipFile(compressed_bin)
        compressed_folder_files = compressed_folder.namelist()

        # test if one and only one file matches the pattern
        files = list(filter(re.compile("produkt").search,
                            compressed_folder_files))

        if len(files) == 0:
            raise ValueError(
                "There is no file matching the pattern: produkt " +
                "in the zip files: \n- " +
                "\n- ".join(compressed_folder_files))
        elif len(files) > 1:
            raise ValueError(
                "There are more than one files matching the " +
                "pattern: produkt\nin the zip file: " +
                str(zip_filepath) +
                "\nonly the first file is returned: " +
                str(files[0]))

        # extract the file from the zip folder and return it as pd.DataFrame
        with compressed_folder.open(files[0]) as f:
            df = pd.read_table(f, sep=";",
                             dtype={"Datum":str, "MESS_DATUM":str},
                               skipinitialspace=True,
                               na_values=[-999, -9999, "####", "#####", "######"])

    elif re.search("derived", zip_filepath):
        df = pd.read_table(f"ftp://{CDC_HOST}/{zip_filepath}",
                             compression="gzip",
                             sep=";",
                             skipinitialspace=True,
                             dtype={"Datum":str, "MESS_DATUM":str},
                             na_values=[-999, -9999, "####", "#####", "######"])
    else:
        raise ImportError("ERROR: No file could be imported, as there is " +
                          "just a setup for observation and derived datas")

    # convert dates to datetime
    for col in ["MESS_DATUM", "Datum"]:
        if col in df.columns:
            df[col] = _dwd_date_parser(df[col])

    return df

def get_dwd_meta(ftp_folder):
    """
    Get the meta file from the ftp_folder on the DWD server.

    Downloads the meta file of a given folder.
    Corrects the meta file of missing files. So if no file for the station is
    in the folder the meta entry gets deleted.
    Reset "von_datum" in meta file if there is a biger gap than max_hole_d.
    Delets entries with less years than min_years.

    Parameters
    ----------
    ftp_folder : str
        The path to the directory where to search for the meta file.
        e.g. "climate_environment/CDC/observations_germany/climate/hourly/precipitation/recent/".

    Returns
    -------
    geopandas.GeoDataFrame
        a GeoDataFrame of the meta file

    """
    # open ftp connection and get list of files in folder
    with ftplib.FTP(CDC_HOST) as ftp:
        ftp.login()

        # get and check the meta_file name
        ftp_files = ftp.nlst(ftp_folder)
        pattern = r".*(?<!_mn4)((_stations_list)|(_Beschreibung_Stationen))+.txt$"
        meta_file = list(filter(re.compile(pattern).match, ftp_files))

    if len(meta_file) == 0:
        log.info(
            f"There is no file matching the pattern '{pattern}'"+
            f"\nin the folder: ftp://{CDC_HOST}/{str(ftp_folder)}")
        return None
    elif len(meta_file) > 1:
        log.info(
            f"There are more than one files matching the pattern: {pattern}" +
            f" in the folder:\nftp://{CDC_HOST}/{str(ftp_folder)}" +
            f"\nonly the first file is returned: {meta_file[0]}")

    # import meta file
    try:
        if re.search("observations", ftp_folder):
            with ftplib.FTP(CDC_HOST) as ftp:
                ftp.login()
                with BytesIO() as bio, StringIO() as sio:
                    ftp.retrbinary("RETR " + meta_file[0], bio.write)
                    sio.write(bio.getvalue().decode("WINDOWS-1252").replace("\r\n", "\n"))
                    colnames = sio.getvalue().split("\n")[0].split()
                    sio.seek(0)
                    meta = pd.read_table(
                        sio,
                        skiprows=2,
                        lineterminator="\n",
                        sep=r"\s{2,}|(?<=\d|\))\s{1}(?=[\w])",  # two or more white spaces or one space after digit and followed by word
                        names=colnames,
                        parse_dates=[col for col in colnames if "datum" in col.lower()],
                        index_col="Stations_id",
                        engine="python")
        elif re.search("derived", ftp_folder):
            meta = pd.read_table("ftp://opendata.dwd.de/" + meta_file[0],
                                 encoding="WINDOWS-1252", sep=";", skiprows=1,
                                 names=["Stations_id", "Stationshoehe",
                                        "geoBreite", "geoLaenge",
                                        "Stationsname", "Bundesland"],
                                 index_col="Stations_id"
                                 )
    except:
        traceback.print_exc()
        print("URL Error: The URL could not be found:\n" +
              "ftp://opendata.dwd.de/" + meta_file[0])
        return None

    try:
        meta = gpd.GeoDataFrame(meta,
                                geometry=gpd.points_from_xy(meta.geoLaenge,
                                                            meta.geoBreite,
                                                            crs="EPSG:4326"))
        meta = meta.drop(["geoLaenge", "geoBreite"], axis=1)
    except:
        traceback.print_exc()
        print("Error while converting DataFrame to GeoDataFrame," +
              " maybe the columns aren't named 'geoLaenge' and geoBreite'" +
              "\nhere is the header of the DataFrame:\n")
        print(meta.head())
        return None

    # delete entries where there is no file in the ftp-folder
    rows_drop = []
    str_ftp_files = str(ftp_files)
    for i, row in meta.iterrows():
        if not (re.search(r"[_\.]" + dwd_id_to_str(i) + r"[_\.]|" +
                          r"[_\.]" + str(i) + r"[_\.]", str_ftp_files)):
            rows_drop.append(i)
    meta = meta.drop(rows_drop)

    # change meta date entries if the file has a different date
    if ("observation" in ftp_folder) \
            and ("bis_datum" and "von_datum" in meta) \
            and ("recent" not in ftp_folder):
        zip_files = list(filter(re.compile(r".+\d+_\d+_\d+_hist.zip").match,
                                ftp_files))
        zip_files.sort()
        zip_files.append(zip_files[0])  # else the last entry won't get tested
        last_sid, last_from_date, last_to_date = None, None, None

        for zip_file in zip_files:
            # get new files dates
            filename = zip_file.split("/")[-1]
            _, kind, sid, from_date, to_date, _ = filename.split("_")
            if kind in ["niedereder"]:
                continue
            from_date = pd.Timestamp(from_date)
            to_date = pd.Timestamp(to_date)
            sid = int(sid)

            # compare with previous file's dates
            if last_sid and (sid == last_sid):
                last_to_date = to_date
            else:
                # compare last values with meta file dates
                if last_sid and (last_sid in meta.index):
                    if last_from_date > meta.loc[last_sid, "von_datum"]:
                        meta.loc[last_sid, "von_datum"] = last_from_date
                    if last_to_date < meta.loc[last_sid, "bis_datum"]:
                        meta.loc[last_sid, "bis_datum"] = last_to_date

                # set values as last values
                last_to_date = to_date
                last_from_date = from_date
                last_sid = sid

    # trim whitespace in string columns
    for dtype, col in zip(meta.dtypes, meta.columns):
        if pd.api.types.is_string_dtype(dtype) and col != "geometry":
            meta[col] = meta[col].str.strip()

    # return
    return meta


# libraries
import logging
import sqlalchemy as sa
from sqlalchemy import text as sqltxt
from functools import cached_property

from ..db.connections import db_engine
from ..utils.dwd import dwd_id_to_str
from ..db.models import MetaPD
from .StationBases import StationPBase, StationCanVirtualBase

# set settings
# ############
__all__ = ["StationPD"]
log = logging.getLogger(__name__)

# class definition
##################
class StationPD(StationPBase, StationCanVirtualBase):
    """A class to work with and download daily precipitation data for one station.

    Those station data are only downloaded to do some quality checks on the 10 minute data.
    Therefor there is no special quality check and richter correction done on this data.
    If you want daily precipitation data, better use the 10 minutes station(StationP) and aggregate to daily values."""

    # common settings
    _MetaModel = MetaPD
    _para = "p_d"
    _para_base = "p"
    _para_long = "daily Precipitation"
    _unit = "mm/day"
    _valid_kinds = {"raw", "filled", "filled_by"}
    _best_kind = "filled"

    # cdc dwd parameters
    _ftp_folder_base = [
        "climate_environment/CDC/observations_germany/climate/daily/kl/",
        "climate_environment/CDC/observations_germany/climate/daily/more_precip/"]
    _cdc_col_names_imp = ["RSK"]
    _db_col_names_imp = ["raw"]

    # timestamp configurations
    _tstp_format_db = "%Y%m%d"
    _tstp_format_human = "%Y-%m-%d"
    _tstp_dtype = "date"
    _interval = "1 day"

    # aggregation
    _min_agg_to = "day"

    # methods from the base class that should not be active for this class
    quality_check = property(doc='(!) Disallowed inherited')
    last_imp_quality_check = property(doc='(!) Disallowed inherited')
    get_corr = property(doc='(!) Disallowed inherited')
    get_adj = property(doc='(!) Disallowed inherited')
    get_qc = property(doc='(!) Disallowed inherited')

    def __init__(self, id, **kwargs):
        super().__init__(id, **kwargs)
        self.id_str = dwd_id_to_str(id)

    def _download_raw(self, zipfiles):
        df_all, max_hist_tstp = super()._download_raw(zipfiles)

        # fill RSK with values from RS if not given
        if "RS" in df_all.columns and "RSK" in df_all.columns:
            mask = df_all["RSK"].isna()
            df_all.loc[mask, "RSK"] = df_all.loc[mask, "RS"]
        elif "RS" in df_all.columns:
            df_all["RSK"] = df_all["RS"]

        return df_all, max_hist_tstp

    @cached_property
    def _table(self):
        return sa.table(
            f"{self.id}_{self._para}",
            sa.column("timestamp", sa.Date),
            sa.column("raw", sa.Integer),
            sa.column("filled", sa.Integer),
            sa.column("filled_by", sa.SmallInteger),
            schema="timeseries")

    @db_engine.deco_create_privilege
    def _create_timeseries_table(self):
        """Create the timeseries table in the DB if it is not yet existing."""
        sql_add_table = '''
            CREATE TABLE IF NOT EXISTS timeseries."{stid}_{para}"  (
                timestamp date PRIMARY KEY,
                raw int4,
                filled int4,
                filled_by int2
            );
        '''.format(stid=self.id, para=self._para)
        with db_engine.connect() as con:
            con.execute(sqltxt(sql_add_table))
            con.commit()




# libraries
import logging

from ..db.connections import db_engine
from .StationsBase import StationsBase
from .StationsP import StationsP

# set settings
# ############
__all__ = ["StationsBaseTET"]
log = logging.getLogger(__name__)

# class definition
##################
class StationsBaseTET(StationsBase):
    @db_engine.deco_update_privilege
    def fillup(self, only_real=False, stids="all", **kwargs):
        # create virtual stations if necessary
        if not only_real:
            meta = self.get_meta(
                infos=["Station_id"], only_real=False)
            meta_p = StationsP().get_meta(
                infos=["Station_id"], only_real=False)
            stids_missing = set(meta_p.index.values) - set(meta.index.values)
            if stids != "all":
                stids_missing = set(stids).intersection(stids_missing)
            for stid in stids_missing:
                self._StationClass(stid) # this creates the virtual station

        super().fillup(only_real=only_real, stids=stids, **kwargs)
